// stringlist functions

typedef struct stringlist
    {
    UINT owner;
    UINT count;
    UINT alloced;
    UINT allocby;
    UINT *strarr;
    } strlist;

strl_initnew(strlist *sl, UINT allocby, UINT owner)
    {
    sl->count = 0;
    sl->allocby = (allocby+15) & 0xF7;
    sl->alloced = 0;
    sl->strarr = NULL;
    sl->owner = owner;
    return 0;
    }

strl_realloc(strlist *sl, unsigned newnum)
    {
    UINT *newarr = (void *) mem_malloc(newnum * 2,sl->owner);
    int i;
    printf("reallocing list from %i (%.4X) to %i (%.4X)",sl->alloced,FP_SEG(sl->strarr),newnum,FP_SEG(newarr));
    if(!newarr)
        return ERR_OUTOFMEMORY;
    for(i=0;i<newnum;i++)
        if(i < sl->alloced)
            newarr[i] = sl->strarr[i];
        else
            newarr[i] = NULL;
    printf("memfreestrlrealloc: OFF=%.4X SEG=%.4X",sl->strarr);
    mem_free((char *) sl->strarr);
    sl->strarr = newarr;
    sl->alloced = newnum;
    return 0;
    }

strl_additem(strlist *sl, char *ele)
    {
    printf("adding %s to sl @ pos %i (alloced:%i)",ele,sl->count,sl->alloced);
    if(sl->alloced <= sl->count)
        if(strl_realloc(sl,sl->count+sl->allocby))
            return ERR_OUTOFMEMORY;
    ele = mem_strnew(ele,sl->owner);
    sl->strarr[sl->count++] = FP_SEG(ele);
    return 0;
    }

#define strl_getitem(sl,index) ((char *) MK_FP(((strlist *) sl)->strarr[index],0))

strl_setitem(strlist *sl, int index, char *newtext)
    {
    printf("memfreestrlsetitem");
    mem_free(strl_getitem(sl,index));
    newtext = mem_strnew(newtext,sl->owner);
    sl->strarr[index] = FP_SEG(newtext);
    }

strl_deleteitem(strlist *sl, int index)
    {
    int i;
    printf("memfreestrldeleteitem");
    mem_free(strl_getitem(sl,index));
    for(i=index;i<(sl->count-1);i++)
        sl->strarr[i] = sl->strarr[i+1];
    sl->strarr[--sl->count] = NULL;
    if(sl->count < (sl->alloced-sl->allocby))
        strl_realloc(sl,sl->count);
    }

strl_clear(strlist *sl)
    {
    int i;
    if(!sl->strarr) return 1;
    printf("for-memfreestrlclear");
    for(i=0;i<sl->count;i++)
        mem_free(strl_getitem(sl,i));
    sl->count = 0;
    sl->alloced = 0;
    printf("memfreestrlclearlist");
    mem_free((char *) sl->strarr);
    sl->strarr = NULL;
    }

strl_free(strlist *sl)
    {
    strl_clear(sl);
    printf("memfreestrlfree");
    mem_free((char *) sl);
    }

void strl_plusminus(strlist *sl, char *op)
    {
    int i;
    if(*op == '+')
        strl_additem(sl,op+1);
    else if(*op++ == '-')
        for(i=0;i<sl->count;i++)
            if(!strcmp(strl_getitem(sl,i),op))
                {
                strl_deleteitem(sl,i);
                return;
                }
    }


strlcompfuncascen(const void *a, const void *b) { return strcmp(MK_FP(*((unsigned *) a),0),MK_FP(*((unsigned *) b),0)); }
strlcompfuncdesc(const void *a, const void *b) { return -strcmp(MK_FP(*((unsigned *) a),0),MK_FP(*((unsigned *) b),0)); }
void strl_sort(strlist *sl, int dodesc)
    {
    qsort(sl->strarr,sl->count,2,dodesc ? &strlcompfuncdesc : &strlcompfuncascen);
    }

void strl_fromsep(strlist *sl, char *data, char sep)
    {
    char a,*p=EINS;
    while(p)
        {
        if((p = strchr(data,sep)) != NULL) a=*p, *p = 0;
        strl_additem(sl,data);
        if(p)
            *p = a,data = p+1;
        }
    }

void strl_swap(strlist *sl, int item1, int item2)
    {
    unsigned x = sl->strarr[item1];
    sl->strarr[item1] = sl->strarr[item2];
    sl->strarr[item2] = x;
    }

strl_additemat(strlist *sl, char *ele, int pos)
    {
    int i;
    printf("addingat %s to sl @ pos %i (alloced:%i)",ele,pos,sl->alloced);
    if(sl->alloced <= sl->count)
        if(strl_realloc(sl,sl->count+sl->allocby))
            return ERR_OUTOFMEMORY;
    ele = mem_strnew(ele,sl->owner);
    for(i=sl->count;i>pos;i--)
        sl->strarr[i] = sl->strarr[i-1];
    sl->strarr[pos] = FP_SEG(ele);
    sl->count++;
    return 0;
    }
